/****************************************************************
*	elfwrap.c: WRAPper for ELF object files.
*
*	``elfwrap'' modifies the symbol/string tables of an
*	object file (a.out and .o) files so that wrapper
*	functions for as-yet unresolved external functions can
*	be written and linked, transparently to the file
*	being wrapped. The technique of binary wrapping is
*	described in the paper by Jon Cargille and Bart Miller.
*
*	This implementation of ``elfwrap'' is specific to
*	object files in the ELF format -- specifically tested
*	on Solaris 2.x
*
*	The usage of elfwrap is:
*		elfwrap [-w] <inpobj> <outobj> <inpsym> <outsym>
*
*	The file <inpobj> is first copied into file <outobj>.
*	Then the symbol and string tables of <outobj> are
*	scanned for an occurence of the symbol name <inpsym>.
*	If such a symbol is found (and if it corresponds to
*	an external (aka global) function, it is replaced
*	with the new symbol named <outsym>. The modified
*	symbol and string tables are written out to the new
*	file <outobj>.
*
*	If the `-w' option is specified on the commandline, then
*	the input symbol is not replaced, but `weakened'. This
*	involves the addition of a new symbol with a WEAK binding
*	to the symbol table. The weak symbol has the name of
*	the output symbol and points to the same object as the
*	given input symbol.
*
*	It is an error if the symbol name <inpsym> is not
*	found in the input object file.
*
*	NOTES on ELF:
*	The ELF library provides a very clean interface for
*	the manipulation of an object or archive file. ELF
*	treats such a file as an abstract Object. Each Object
*	consists of sub-Objects which correspond to different
*	sections (or segments) of the disk file.
*
*	Each section (or sub-Object) has a section header
*	and section data. Methods are provided to manipulate
*	section headers and data -- the modification of actual
*	data within each section is left to the application
*	program using the ELF library.
*
*	NOTES on SYMBOL/STRING table:
*	A symbol table consists of a short summary of some/all
*	entities of a program -- variable names, internal &
*	external functions, types, filenames etc. The symbol
*	table entry is pure binary -- and indicates what kind
*	of a symbol is being described, what its value is and
*	where in the object file it can be found.
*
*	The actual text-name of the symbol is found in the
*	string table. There exists a pointer in the symbol
*	table entry into the string table -- for those
*	symbols that have names. Unresolved external references
*	are finally resolved by a linker using string-names
*	that are matched with the string table. Hence, binary
*	wrapping makes use of this string-symbol association
*	to transparently add front-ends to as-yet unresolved
*	functions.
*
*	Ideally, the pointer from a symbol table entry into the
*	string table should be an entry index, or there should
*	be a provision for some kind of a permuted reverse index
*	from the string table into the symbol table. This does
*	not exist in any current symbol table format. The
*	pointer is usually a file or memory offset relative to
*	the start of the file, or start of a section.
*
*	Therefore changes to a single entry in the string table
*	affects symbol table pointers for all entries pointing
*	beyond this string. The current implementation of
*	``elfwrap'' assumes that the symbol table entries are
*	sorted based on the value of their pointer into the
*	string table. This is true for all symbol tables formats
*	generated by compiler suites.
****************************************************************/





/****************************************************************
*	Include header files.
****************************************************************/

#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<libelf.h>





/****************************************************************
*	Define symbolic constants.
****************************************************************/

#define		F_OKAY		0
#define		F_ERR		-1

#define		WEAKEN_OPT	"-w"
#define		NUM_ARGS_REQD	4

#define		F_MODE		0666

#ifndef		BUFSIZ
#define		BUFSIZ		1024
#endif

#define		SYMTAB_NAME	".symtab"
#define		STRTAB_NAME	".strtab"





/****************************************************************
*	Define macros.
****************************************************************/

#define		basename(path, base)				\
		if (!(base = (char *)strrchr(path, '/')))	\
			base = path;				\
		else						\
			base++;

#define		usage()						\
		fprintf(stderr, "%s: %s %s %s %s %s %s\n",	\
			"Usage", progname, "[-wg]",		\
			"<inpobj>", "<outobj>",			\
			"<inpsym>", "<outsym>");





/****************************************************************
*	Define global variables.
****************************************************************/

static char*		progname = "a.out";

static int		wflag = 0;

static char*		inpobj;
static char*		outobj;
static char*		inpsym;
static char*		outsym;

static Elf*		elfd;
static Elf32_Ehdr*	ehdrd;

static int		symsec;
static Elf_Scn*		symscnd;
static Elf32_Shdr*	symshdr;
static Elf_Data*	symdata;

static int		strsec;
static Elf_Scn*		strscnd;
static Elf32_Shdr*	strshdr;
static Elf_Data*	strdata;

static int		nsyms;
static Elf32_Sym*	syms;





/****************************************************************
*	elf_perror(char* str)
*
*	Use the ELF errno and error message functions to print
*	out error messages arising from ELF functions.
*
*	elf_errno() clears the value of the ELF error number
*	as soon as it is called once. Hence, elf_perror is
*	not idempotent.
****************************************************************/

static void
elf_perror(str)
char*	str;
{
	char*	msg;


	if (!(msg = (char *) elf_errmsg(elf_errno()))) msg = "";
	fprintf(stderr, "%s: %s\n", str, msg);
	return;
}





/****************************************************************
*	int copy_file(char* inp, char* out)
*
*	Copy the contents of file "inp" into "out". The output
*	file is created anew.
****************************************************************/

static int
copy_file(inp, out)
char*	inp;
char*	out;
{
	int	ifd;
	int	ofd;
	ssize_t	n;
	char	buf[BUFSIZ];


	if ((ifd = open(inp, O_RDONLY)) < 0) {
		perror(inp);
		return F_ERR;
	}
	if ((ofd = creat(out, F_MODE)) < 0) {
		perror(out);
		return F_ERR;
	}

	while ((n = read(ifd, (void *) buf, BUFSIZ)) > 0) {
		if (write(ofd, (void *) buf, (size_t) n) != n) {
			perror(out);
			return F_ERR;
		}
	}

	if (n < 0) {
		perror(inp);
		return F_ERR;
	}

	(void) close(ifd);
	(void) close(ofd);
	return F_OKAY;
}





/****************************************************************
*	int locate_symstr(void)
*
*	ELF object files have a special string table which is
*	simply a collection of the names of the other sections
*	in the object file. This section index is given by the
*	"e_shstrndx" field of an elf execution header.
*
*	The actual section names are data in this section. We
*	need to scan this section for the special sequences
*	".symtab" and ".strtab" which identify the symbol and
*	string table sections rspy.
*
*	The index of these names in the section string table
*	corresponds to the index of the actual sections in the
*	object file. These indices can be provided as arguments
*	to the elf_getscn() function to retrieve the actual
*	sections (headers/data).
*
*	This function sets "symsec" and "strsec" to be the
*	indices of the symbol and string table rspy. If either
*	of these sections is not locatable via the section string
*	table, an error is returned.
*
*	This function uses the global values of the ELF and
*	Execution Header descriptors and sets the values of
*	the global symbol and string table indices.
****************************************************************/

static int
locate_symstr()
{
	Elf_Scn*	shstrscnd;
	Elf_Data*	shstrdata;
	char*		secnames;
	u_int		i;


	symsec = -1;
	strsec = -1;

	if (!(shstrscnd = elf_getscn(elfd, ehdrd->e_shstrndx))) {
		elf_perror(outobj);
		return F_ERR;
	}
	if (!(shstrdata = elf_getdata(shstrscnd, 0))) {
		elf_perror(outobj);
		return F_ERR;
	}
	secnames = (char *) shstrdata->d_buf;

	i = 0;
	while (i < ehdrd->e_shnum) {
		if (strcmp(secnames, SYMTAB_NAME) == 0) symsec = i;
		if (strcmp(secnames, STRTAB_NAME) == 0) strsec = i;

		secnames += (1+strlen(secnames));
		i++;
	}

	if (symsec == -1) {
		fprintf(stderr, "%s: no symbol table in %s\n", progname, outobj);
		return F_ERR;
	}
	else if (strsec == -1) {
		fprintf(stderr, "%s: no string table in %s\n", progname, outobj);
		return F_ERR;
	}
	return F_OKAY;
}





/****************************************************************
*	int load_tables(void)
*
*	Once the indices of the symbol and string tables have
*	been determined, this function allocates section
*	descriptors for these two sections and reads in the
*	section headers and section data for both these sections
****************************************************************/

static int
load_tables()
{
	if (!(symscnd = elf_getscn(elfd, symsec))) {
		elf_perror(outobj);
		return F_ERR;
	}
	if (!(strscnd = elf_getscn(elfd, strsec))) {
		elf_perror(outobj);
		return F_ERR;
	}

	if (!(symshdr = elf32_getshdr(symscnd))) {
		elf_perror(outobj);
		return F_ERR;
	}
	if (!(strshdr = elf32_getshdr(strscnd))) {
		elf_perror(outobj);
		return F_ERR;
	}

	if (symshdr->sh_type != SHT_SYMTAB) return F_ERR;
	if (strshdr->sh_type != SHT_STRTAB) return F_ERR;

	if (!(symdata = elf_getdata(symscnd, 0))) {
		elf_perror(outobj);
		return F_ERR;
	}
	if (!(strdata = elf_getdata(strscnd, 0))) {
		elf_perror(outobj);
		return F_ERR;
	}

	return F_OKAY;
}





/****************************************************************
*	int init_elf(void)
*
*	o Create an ELF descriptor for the output object file
*	o Read in file headers
*	o Verify type and class of the object file
*	o Load the symbol and string tables into memory
****************************************************************/

static int
init_elf()
{
	int		fd;
	Elf_Cmd		elfcmd;
	Elf_Kind	elfkind;


	if ((fd = open(outobj, O_RDWR)) < 0) {
		perror(outobj);
		return F_ERR;
	}

	(void) elf_version(EV_CURRENT);
	(void) elf_errno();
	elfcmd = ELF_C_RDWR;
	if (!(elfd = elf_begin(fd, elfcmd, (Elf *)0))) {
		elf_perror(outobj);
		return F_ERR;
	}

	elfkind = elf_kind(elfd);
	if (elfkind != ELF_K_ELF) {
		fprintf(stderr, "%s: %s is *not* in ELF format\n",
			progname, outobj);
		return F_ERR;
	}

	if (!(ehdrd = elf32_getehdr(elfd))) {
		elf_perror(outobj);
		return F_ERR;
	}
	if (ehdrd->e_ident[EI_CLASS] != ELFCLASS32) {
		fprintf(stderr, "%s: %s not in 32-bit class\n", progname, outobj);
		return F_ERR;
	}

	if (locate_symstr() != F_OKAY) {
		fprintf(stderr, "%s: unable to locate str/sym tables\n", progname);
		fprintf(stderr, "%s: cannot process %s\n", progname, outobj);
		return F_ERR;
	}

	if (load_tables() != F_OKAY) {
		fprintf(stderr, "%s: unable to load str/sym tables\n", progname);
		fprintf(stderr, "%s: cannot process %s\n", progname, outobj);
		return F_ERR;
	}

	return F_OKAY;
}





/****************************************************************
*	int adjust_symbol(u_int i)
*
*	This is the trickiest function of ``elfwrap''.
*	The input parameter "i" indicates that symbol table entry
*	"i" has a name matching the input symbol name and needs
*	to be replaced with the new output symbol name.
*
*	Since, the input and output symbol names can be of
*	different lengths, the difference is computed. If the
*	new length is larger than the old one, space is allocated
*	for the new string table (the symbol table size does not
*	ever change during binary wrapping, only the contents of
*	the entries changes).
*
*	Once the new string table has been allocated, all symbols
*	preceeding the changing symbol can be copied as is. The
*	old symbol is then replaced by the new one and the rest
*	of the symbols following the old symbol in the old table
*	are copied as is to the new table. This work is done by
*	the three memcpy functions. Then the string section
*	data is updated to point to the new string table.
*
*	The final loop updates the symbol table pointers into
*	the string table. If there is a length change in the
*	symbol name, all symbol table pointers will either
*	shift up or down by this length difference. This is
*	where the ordering of the symbol table entries is made
*	use of in updating only symbols "i" and beyond.
****************************************************************/

static int
adjust_symbol(i)
u_int	i;
{
	u_int	j;
	int	sizediff;
	u_int	newsize;
	void*	buf;
	void*	src;
	void*	dst;
	u_int	nbytes;


	sizediff = strlen(outsym) - strlen(inpsym);
	newsize = strdata->d_size + sizediff;

	if (!(buf = (void *) malloc(newsize*sizeof(char)))) {
		perror("malloc");
		return F_ERR;
	}

	src = (void *) strdata->d_buf;
	dst = (void *) buf;
	nbytes = syms[i].st_name - 1;
	if (memcpy(dst, src, nbytes) != dst) {
		perror("memcpy");
		return F_ERR;
	}

	src = (void *) outsym;
	dst = (void *) ((char *) buf + syms[i].st_name);
	nbytes = strlen(outsym)+1;
	if (memcpy(dst, src, nbytes) != dst) {
		perror("memcpy");
		return F_ERR;
	}

	src = (void *) ((char *) strdata->d_buf + syms[i+1].st_name);
	dst = (void *) ((char *) buf + syms[i].st_name + strlen(outsym) + 1);
	nbytes = strdata->d_size - syms[i+1].st_name;
	if (memcpy(dst, src, nbytes) != dst) {
		perror("memcpy");
		return F_ERR;
	}

	strdata->d_buf = buf;
	strdata->d_size = newsize;

	for (j = i+1; j < nsyms; j++)
		syms[j].st_name += sizediff;

	return F_OKAY;
}





/****************************************************************
*	int weaken_symbol(u_int i)
*
*	This function adds a new entry to the symbol table,
*	without changing the string table. The symbol table
*	entry is indexed by the input parameter `i' and the
*	new entry is created by extending the symbol table.
*
*	The rest of the symbol table entries are copied as is
*	and the size and buffer in the symbol table data are
*	changed appropriately.
*
*	Once the new symbol table entry is created, expand
*	the string table to accomodate the new symbol name.
****************************************************************/

static int
weaken_symbol(i)
u_int	i;
{
	Elf32_Sym*	newsyms;
	u_int		newnsyms;
	u_int		j;
	u_int		newsize;
	void*		buf;
	void*		src;
	void*		dst;
	u_int		nbytes;


	symdata->d_size = (nsyms+1) * sizeof (Elf32_Sym);
	if (!(newsyms = (Elf32_Sym *) malloc(symdata->d_size))) {
		perror("malloc");
		return F_ERR;
	}

	for (j = 0; j < nsyms; j++) {
		newsyms[j].st_name  = syms[j].st_name;
		newsyms[j].st_value = syms[j].st_value;
		newsyms[j].st_size  = syms[j].st_size;
		newsyms[j].st_info  = syms[j].st_info;
		newsyms[j].st_other = syms[j].st_other;
		newsyms[j].st_shndx = syms[j].st_shndx;
	}

	newsize = strdata->d_size + strlen(outsym) + 1;
	if (!(buf = (void *) malloc(newsize*sizeof(char)))) {
		perror("malloc");
		return F_ERR;
	}

	src = (void *) strdata->d_buf;
	dst = (void *) buf;
	nbytes = strdata->d_size;
	if (memcpy(dst, src, nbytes) != dst) {
		perror("memcpy");
		return F_ERR;
	}

	src = (void *) outsym;
	dst = (void *) ((char *) buf + strdata->d_size);
	nbytes = strlen(outsym)+1;
	if (memcpy(dst, src, nbytes) != dst) {
		perror("memcpy");
		return F_ERR;
	}

	newsyms[nsyms].st_name  = strdata->d_size;
	newsyms[nsyms].st_value = syms[i].st_value;
	newsyms[nsyms].st_size  = syms[i].st_size;
	newsyms[nsyms].st_info  = ELF32_ST_INFO(STB_WEAK,
				  ELF32_ST_TYPE(syms[i].st_info));
	newsyms[nsyms].st_other = syms[i].st_other;
	newsyms[nsyms].st_shndx = syms[i].st_shndx;

	strdata->d_buf  = buf;
	strdata->d_size = newsize;
	symdata->d_buf  = (void *) newsyms;

	return F_OKAY;
}





/****************************************************************
*	int locate_sym(char* str)
*
*	This function scans the string table for the specified
*	string. If the string is found, it is matched up with
*	the symbol table and the index of the name in the
*	symbol table is returned.
*
*	Only GLOBAL/WEAK binding and NOTYPE/FUNC types are
*	searched for. If the name is not found, F_ERR is
*	returned.
*
*	The global variables `syms' and `nsyms' are used
*	to scan the symbol table.
****************************************************************/

static int
locate_sym(str)
char*	str;
{
	u_int	i;
	u_int	symbind;
	u_int	symtype;
	char*	name;


	for (i = 0; i < nsyms; i++) {
		symbind = ELF32_ST_BIND(syms[i].st_info);
		symtype = ELF32_ST_TYPE(syms[i].st_info);

		if ((symbind != STB_GLOBAL) && (symbind != STB_WEAK))
			continue;
		if ((symtype != STT_NOTYPE) && (symtype != STT_FUNC))
			continue;
		
		name = (((char *) strdata->d_buf) + syms[i].st_name);
		if (strcmp(name, str) == 0)
			return i;
	}

	return F_ERR;
}





/****************************************************************
*	int wrap_symbol()
*
*	This function walks thru the symbol table entries.
*	Each entry is checked for its type. If an entry is a
*	global (or external) and is a function/unknown (since
*	library functions are defined as unknown by the assembler
*	we need this hack), only then is the string table
*	entry accessed.
*
*	If the symbol name matches the input symbol, the symbol
*	table and string table are adjusted to accomodate the
*	changed symbol.
****************************************************************/

static int
wrap_symbol()
{
	int	in, out;
	int	ret;


	nsyms = symdata->d_size / sizeof(Elf32_Sym);
	syms  = (Elf32_Sym *) symdata->d_buf;

	if ((in = locate_sym(inpsym)) == F_ERR) {
		fprintf(stderr, "%s: symbol %s not found in %s\n",
			progname, inpsym, inpobj);
		return F_OKAY;
	}

	if (wflag) {
		out = locate_sym(outsym);
		if (out != F_ERR) {
			fprintf(stderr, "%s: symbol `%s' already present\n",
				progname, outsym);
			return F_OKAY;
		}
		return weaken_symbol(in);
	}
	else {
		return adjust_symbol(in);
	}
}





/****************************************************************
*	int update_elf()
*
*	All modifications to ELF objects are made in memory.
*	To update the file data, the modified sections are marked
*	as dirty -- here, the whole file is marked as dirty.
*
*	The close(creat()) idiom specified in the man pages does
*	not seem reliable -- especially for large files, when the
*	ELF libraries do not store the whole object in main
*	memory, but need to access the disk.
*
*	Since the ELF libraries truncate the file properly to
*	size, the direct update is okay.
****************************************************************/

static int
update_elf()
{
	if (elf_flagelf(elfd, ELF_C_SET, ELF_F_DIRTY) == 0) {
		elf_perror(outobj);
		return F_ERR;
	}

	if (elf_update(elfd, ELF_C_WRITE) == -1) {
		elf_perror(outobj);
		return F_ERR;
	}
	if (elf_end(elfd) == -1) {
		elf_perror(outobj);
		return F_ERR;
	}

	return F_OKAY;
}





/****************************************************************
*	int elfwrap(void)
*
*	o Creates a copy of the output object file from the
*	  input.
*	o Read in the ELF data structures and symbol/string
*	  tables.
*	o Search the tables and modify the specified input symbol
*	  with the new output symbol.
*	o Update the output file from the memory structures
****************************************************************/

static int
elfwrap()
{
	if (copy_file(inpobj, outobj) != F_OKAY)
		return F_ERR;

	if (init_elf() != F_OKAY)
		return F_ERR;

	if (wrap_symbol() != F_OKAY)
		return F_ERR;

	if (update_elf() != F_OKAY)
		return F_ERR;

	return F_OKAY;
}





/****************************************************************
*	int main(int argc, char** argv)
*
*	o Check for proper command invocation
*	o Collect command line arguments
*	o Invoke binary wrapper
****************************************************************/

int
main(argc, argv)
int	argc;
char**	argv;
{
	basename(*argv, progname); argv++; argc--;

	if ((argc != 0) && (strcmp(*argv, WEAKEN_OPT) == 0)) {
		wflag++;
		argv++; argc--;
	}
	if (argc != NUM_ARGS_REQD) {
		usage();
		exit(EXIT_FAILURE);
	}

	inpobj = *argv; argv++; argc--;
	outobj = *argv; argv++; argc--;
	inpsym = *argv; argv++; argc--;
	outsym = *argv; argv++; argc--;

	if (elfwrap() == F_ERR) {
		fprintf(stderr, "%s: unable to wrap %s into %s\n",
			progname, inpobj, outobj);
		exit(EXIT_FAILURE);
	}

	return EXIT_SUCCESS;
}

