/************************************************************************
*									*
*	binwrap.c: A object file wrapper for instrumenting libraries	*
*									*
*	Standard language libraries are generally archives of object	*
*	files. The binary wrapper modifies the symbol table entries of	*
*	an object file so that a call to a standard library function	*
*	(or a system call) by a program now could refer to a special	*
*	instrumentation function. This function could duplicate the	*
*	functionality of the original function or could call the	*
*	original AND execute additional code.				*
*									*
*	This program for a binary wrapper is specific to the format of	*
*	the symbol/string tables for the Sun-{2,3,4} architectures.	*
*	Wrappers for different architectures will be different and may	*
*	need to be re-written from scratch if the object file format is	*
*	sufficiently different from the one assumed here.		*
*									*
************************************************************************/





/************************************************************************
*									*
*	Include header files.						*
*									*
************************************************************************/

#include	<stdio.h>
#include	<string.h>
#include	<memory.h>

#include	<sys/types.h>
#include	<unistd.h>
#include	<fcntl.h>

#include	<a.out.h>
#include	<stab.h>





/************************************************************************
*									*
*	Define symbolic constants.					*
*									*
************************************************************************/

#define		F_OK		0
#define		F_ERR		-1
#define		F_EXIT		1

#ifndef		FALSE
#define		FALSE		0
#endif
#ifndef		TRUE
#define		TRUE		1
#endif

#define		NUM_ARGS	4
#define		OBJ_MODE	0666

#define		EXEC_SIZE	(sizeof(struct exec))
#define		NLIST_SIZE	(sizeof(struct nlist))





/************************************************************************
*									*
*	Define macros.							*
*									*
************************************************************************/

#define		Usage		fprintf(stderr, "Usage: %s %s %s %s\n", \
					progName, \
					"[-t]", \
					"<inp-obj> <out-obj>", \
					"<inp-sym> <out-sym>")





/************************************************************************
*									*
*	Define global variables.					*
*									*
************************************************************************/

static	char	*progName;
static	int	textOnly = FALSE;
static	char	*inpObj, *outObj;





/************************************************************************
*									*
*	ReadExec():							*
*		o Seek to the start of the specified file		*
*		o Read the "exec" structure				*
*		o Check for errors in seek/read				*
*									*
*	The "exec" header is found at the start of all objects (i.e.	*
*	files produced as the output of "as" and "ld") under SunOs for	*
*	Sun-{2,3,4}. Refer "man a.out" for more information on "exec".	*
*									*
*	The magic number of the input header is checked for validity.	*
*	No other checks are made (to see if the segment sizes are legal	*
*	or not). All such checks are the responsibility of the calling	*
*	function.							*
*									*
************************************************************************/

int	ReadExec(fd, execPtr, file)
int		fd;
struct exec	*execPtr;
char		*file;
{
	/* Declare variables */
	int	nbytes;


	/* Check for null pointers */
	if (execPtr == (struct exec *)NULL)
	{
		fprintf(stderr, "%s: cannot read null \"struct exec\"\n",
			progName);
		return F_ERR;
	}

	/* Seek to the start of the file */
	if (lseek(fd, (off_t)0, SEEK_SET) < 0)
	{
		fprintf(stderr, "%s: seek error (%s, %d, %d)\n",
			progName, file, fd, 0);
		perror("lseek");
		return F_ERR;
	}

	/* Read in the exec header */
	nbytes = read(fd, (char *)execPtr, EXEC_SIZE);
	if (nbytes < 0)
	{
		fprintf(stderr, "%s: read error (%s, %d)\n",
			progName, file, fd);
		perror("read");
		return F_ERR;
	}
	else if (nbytes < EXEC_SIZE)
	{
		fprintf(stderr, "%s: partial read (%s, %d)\n",
			progName, file, fd);
		perror("read");
		return F_ERR;
	}

	/* Check magic number */
	if (N_BADMAG(*execPtr))
	{
		fprintf(stderr, "%s: bad magic (%s, %d)\n",
			progName, file, fd);
		return F_ERR;
	}

	/* Read successful, return */
	return F_OK;
}





/************************************************************************
*									*
*	WriteExec():							*
*		o Seek to the start of the specified file		*
*		o Write the specified "exec" structure, overwriting any	*
*		  previous contents of the file.			*
*		o Check for errors in seek/write			*
*									*
*	The "exec" header is found at the start of all objects (i.e.	*
*	files produced as the output of "as" and "ld") under SunOs for	*
*	Sun-{2,3,4}. Refer "man a.out" for more information on "exec".	*
*									*
*	No checks are made to see if the "exec" header provided is	*
*	valid or not. It is the responsibility of the calling function	*
*	to provide a legal header.					*
*									*
************************************************************************/

int	WriteExec(fd, execPtr, file)
int		fd;
struct exec	*execPtr;
char		*file;
{
	/* Declare variables */
	int	nbytes;


	/* Check for null pointers */
	if (execPtr == (struct exec *)NULL)
	{
		fprintf(stderr, "%s: cannot write null \"struct exec\"\n",
			progName);
		return F_ERR;
	}

	/* Seek to the start of the file */
	if (lseek(fd, (off_t)0, SEEK_SET) < 0)
	{
		fprintf(stderr, "%s: seek error (%s, %d, %d)\n",
			progName, file, fd, 0);
		perror("lseek");
		return F_ERR;
	}

	/* Write the specified exec structure */
	nbytes = write(fd, (char *)execPtr, EXEC_SIZE);
	if (nbytes < 0)
	{
		fprintf(stderr, "%s: write error (%s, %d)\n",
			progName, file, fd);
		perror("write");
		return F_ERR;
	}
	else if (nbytes < EXEC_SIZE)
	{
		fprintf(stderr, "%s: partial write (%s, %d)\n",
			progName, file, fd);
		perror("write");
		return F_ERR;
	}

	/* Write successful, return */
	return F_OK;
}





/************************************************************************
*									*
*	CopyStdInfo():							*
*		o Compute the sizes of the standard segments		*
*		o Create a "one-shot" buffer for this data		*
*		o Read text, data and reloc. info from input file	*
*		o Write this data to output file			*
*		o Check for errors in seek/read/write			*
*									*
*	An object files program text, data and relocation information	*
*	does not change when symbol/string table entries are modified.	*
*	Moreover, these segments are stored BEFORE the symbol/string	*
*	table entries in the object files, hence any size changes to	*
*	the tables do not affect the file offsets of these segments.	*
*	Hence, they are copied to unmodified to the output file. If the	*
*	relocation table information needs to be modified for some	*
*	reason (wrap recursive functions ??), this function will need	*
*	to be modified appropriately.					*
*									*
************************************************************************/

int	CopyStdInfo(inpFd, outFd, execPtr)
int		inpFd, outFd;
struct exec	*execPtr;
{
	/* Declare variables */
	int	start, end;
	int	bufSize, nbytes;
	char	*buf;


	/* Check for null pointers */
	if (execPtr == (struct exec *)NULL)
	{
		fprintf(stderr, "%s: cannot process null \"struct exec\"\n",
			progName);
		return F_ERR;
	}

	/* Get size of unchanged portions, initialize buffer */
	start = N_TXTOFF(*execPtr);
	end = N_SYMOFF(*execPtr);
	bufSize = end - start;
	if ((buf = (char *)malloc(bufSize*sizeof(char))) == (char *)NULL)
	{
		fprintf(stderr, "%s: memory allocation, %d bytes\n",
			progName, (bufSize*sizeof(char)));
		perror("malloc");
		return F_ERR;
	}

	/* Read text, data and reloc info from input file */
	if (lseek(inpFd, (off_t)start, SEEK_SET) < 0)
	{
		fprintf(stderr, "%s: seek error (%s, %d, %d)\n",
			progName, inpObj, inpFd, start);
		perror("lseek");
		return F_ERR;
	}
	nbytes = read(inpFd, buf, bufSize);
	if (nbytes < 0)
	{
		fprintf(stderr, "%s: read error (%s, %d)\n",
			progName, inpObj, inpFd);
		perror("read");
		return F_ERR;
	}
	else if (nbytes < bufSize)
	{
		fprintf(stderr, "%s: partial read (%s, %d)\n",
			progName, inpObj, inpFd);
		perror("partial");
		return F_ERR;
	}

	/* Write info to the output file */
	if (lseek(outFd, (off_t)start, SEEK_SET) < 0)
	{
		fprintf(stderr, "%s: seek error (%s, %d, %d)\n",
			progName, outObj, outFd, start);
		perror("lseek");
		return F_ERR;
	}
	nbytes = write(outFd, buf, bufSize);
	if (nbytes < 0)
	{
		fprintf(stderr, "%s: write error (%s, %d)\n",
			progName, outObj, outFd);
		perror("write");
		return F_ERR;
	}
	else if (nbytes < bufSize)
	{
		fprintf(stderr, "%s: partial write (%s, %d)\n",
			progName, outObj, outFd);
		perror("write");
		return F_ERR;
	}

	/* Free buffer */
	(void) free(buf);

	/* Standard information copied, return */
	return F_OK;
}





/************************************************************************
*									*
*	CopyTables():							*
*		o Compute size of the symbol/string tables		*
*		o Allocate memory for the tables			*
*		o Read in the symbol and string tables into memory	*
*		o Scan the string table and replace the old symbol name	*
*		  with the new name, adjust table size			*
*		o Write the new symbol/string table and table sizes to	*
*		  new object file					*
*		o Check for seek/read/write errors			*
*									*
*	The symbol table consists of a set of "nlist" structures, each	*
*	describing the file-pointer into the string table area of the	*
*	object file where the name of the symbol is stored and entries	*
*	describing the type of the symbol.				*
*									*
*	Modifying a string table entry means that the size of the	*
*	string table will change. This means that the symbol table file	*
*	pointers will have to be adjusted. The symbol table entries are	*
*	in general generated in order, this is implicitly assumed in	*
*	this function. If the symbol table entries are not sorted by	*
*	file position of the corresponding strings, a sorting needs to	*
*	be done before the old symbol is replaced by the new one.	*
*									*
************************************************************************/

int	CopyTables(inpFd, outFd, inpSym, outSym, inpX, outX)
int		inpFd, outFd;
char		*inpSym, *outSym;
struct exec	*inpX, *outX;
{
	/* Declare variables */
	int		nsyms;
	struct nlist	*symtab;
	char		**strtab;
	int		nbytes;
	int		i;
	long		strptr;


	/* Compute the number of symbols in input file */
	nsyms = inpX->a_syms / sizeof(struct nlist);
	if (nsyms == 0)
	{
		fprintf(stderr, "%s: %s has no symbols\n",
			progName, inpObj);
		return F_OK;
	}

	/* Allocate space for symbol and string tables */
	if ((symtab = (struct nlist *)malloc(nsyms*NLIST_SIZE)) ==
		(struct nlist *)NULL)
	{
		fprintf(stderr, "%s: memory allocation, %d bytes\n",
			progName, (nsyms*NLIST_SIZE));
		perror("malloc");
		return F_ERR;
	}
	if ((strtab = (char **)malloc(nsyms*sizeof(char *))) == (char **)NULL)
	{
		fprintf(stderr, "%s: memory allocation, %d bytes\n",
			progName, (nsyms*sizeof(char *)));
		perror("malloc");
		return F_ERR;
	}

	/* Read in the symbol table */
	if (lseek(inpFd, (off_t)N_SYMOFF(*inpX), SEEK_SET) < 0)
	{
		fprintf(stderr, "%s: seek error (%s, %d, %d)\n",
			progName, inpObj, inpFd, N_SYMOFF(*inpX));
		perror("lseek");
		return F_ERR;
	}
	nbytes = read(inpFd, (char *)symtab, inpX->a_syms);
	if (nbytes < 0)
	{
		fprintf(stderr, "%s: read error (%s, %d)\n",
			progName, inpObj, inpFd);
		perror("read");
		return F_ERR;
	}
	else if (nbytes < inpX->a_syms)
	{
		fprintf(stderr, "%s: partial read (%s, %d)\n",
			progName, inpObj, inpFd);
		perror("read");
		return F_ERR;
	}

	/* Read in the string table */
	for (i = 0; i < nsyms; i++)
	{
		/* Declare variables */
		char	ch;
		int	size;
		off_t	pos;


		/* Check for invalid string pointer (special symbol) */
		if (symtab[i].n_un.n_strx == 0)
		{
			strtab[i] = (char *)NULL;
			continue;
		}

		/* Get symbol size */
		pos = (off_t)(N_STROFF(*inpX) + symtab[i].n_un.n_strx);
		if (lseek(inpFd, pos, SEEK_SET) < 0)
		{
			fprintf(stderr, "%s: seek error (%s, %d, %d)\n",
				progName, inpObj, inpFd, N_STROFF(*inpX));
			perror("lseek");
			return F_ERR;
		}
		size = 0;
		do
		{
			nbytes = read(inpFd, (char *)&ch, sizeof(char));
			if (nbytes < 0)
			{
				fprintf(stderr, "%s: read error (%s, %d)\n",
					progName, inpObj, inpFd);
				return F_ERR;
			}
			else if (nbytes == 0)
			{
				fprintf(stderr, "%s: unexpected eof in string table %s\n",
					progName, inpObj);
				return F_ERR;
			}
			size++;
		}
		while (ch != '\0');

		/* Read symbol string */
		if (lseek(inpFd, pos, SEEK_SET) < 0)
		{
			fprintf(stderr, "%s: seek error (%s, %d, %d)\n",
				progName, inpObj, inpFd, pos);
			perror("lseek");
			return F_ERR;
		}
		if ((strtab[i] = (char *)malloc(size)) == (char *)NULL)
		{
			fprintf(stderr, "%s: memory allocation, %d bytes\n",
				progName, size);
			perror("malloc");
			return F_ERR;
		}
		nbytes = read(inpFd, strtab[i], size);
		if (nbytes < 0)
		{
			fprintf(stderr, "%s: read error (%s, %d)\n",
				progName, inpObj, inpFd);
			perror("read");
			return F_ERR;
		}
		else if (nbytes < size)
		{
			fprintf(stderr, "%s: partial read (%s, %d)\n",
				progName, inpObj, inpFd);
			perror("read");
			return F_ERR;
		}
	}

	/* Check each symbol for modifications */
	strptr = 4;
	for (i = 0; i < nsyms; i++)
	{
		/* Declare variables */
		char	*str;


		/* Ignore null symbols */
		if (strtab[i] == (char *)NULL)
			continue;

		/* Modify string table pointer, due to previous replacements */
		symtab[i].n_un.n_strx = strptr;

		/* Ignore non-external symbols */
		if (!(symtab[i].n_type & N_EXT))
			goto updatePtr;

		/* Check for text symbol and option */
		if (textOnly && !((symtab[i].n_type&N_TYPE)&N_TEXT))
			goto updatePtr;

		/* Ignore non-matching strings */
		str = strtab[i]; str++;
		if (strcmp(inpSym, str) != 0)
			goto updatePtr;

		/* Replace old symbol with new in memory */
		(void) free(strtab[i]);
		if ((strtab[i] = (char *)malloc(strlen(outSym)+2)) == (char *)NULL)
		{
			fprintf(stderr, "%s: memory allocation, %d bytes\n",
				progName, (strlen(outSym)+2));
			perror("malloc");
			return F_ERR;
		}
		strcpy(strtab[i], "_"); str = strtab[i]; str++;
		(void) memcpy(str, outSym, (strlen(outSym)+1));

		/* Update string table pointer */
updatePtr:
		strptr += strlen(strtab[i])+1;
	}

	/* Write out the modified symbol table */
	if (lseek(outFd, (off_t)N_SYMOFF(*inpX), SEEK_SET) < 0)
	{
		fprintf(stderr, "%s: seek error (%s, %d, %d)\n",
			progName, outObj, outFd, N_SYMOFF(*inpX));
		perror("lseek");
		return F_ERR;
	}
	nbytes = write(outFd, (char *)symtab, inpX->a_syms);
	if (nbytes < 0)
	{
		fprintf(stderr, "%s: write error (%s, %d)\n",
			progName, outObj, outFd);
		perror("write");
		return F_ERR;
	}
	else if (nbytes < inpX->a_syms)
	{
		fprintf(stderr, "%s: partial write (%s, %d)\n",
			progName, outObj, outFd);
		perror("write");
		return F_ERR;
	}

	/* Write size of the string table */
	nbytes = write(outFd, (char *)&strptr, sizeof(strptr));
	if (nbytes < 0)
	{
		fprintf(stderr, "%s: write error (%s, %d)\n",
			progName, outObj, outFd);
		perror("write");
		return F_ERR;
	}
	else if (nbytes < sizeof(strptr))
	{
		fprintf(stderr, "%s: partial write (%s, %d)\n",
			progName, outObj, outFd);
		perror("write");
		return F_ERR;
	}

	/* Write out the modified string table */
	for (i = 0; i < nsyms; i++)
	{
		/* Declare variables */
		int	size;


		/* Ignore null symbols */
		if (strtab[i] == (char *)NULL)
			continue;
		
		/* Write out the string to the output file */
		size = strlen(strtab[i])+1;
		nbytes = write(outFd, strtab[i], size);
		if (nbytes < 0)
		{
			fprintf(stderr, "%s: write error (%s, %d)\n",
				progName, outObj, outFd);
			perror("write");
			return F_ERR;
		}
		else if (nbytes < size)
		{
			fprintf(stderr, "%s: partial write (%s, %d)\n",
				progName, outObj, outFd);
			perror("write");
			return F_ERR;
		}
	}

	/* Free buffers */
	(void) free(symtab);
	for (i = 0; i < nsyms; i++)
		(void) free(strtab[i]);
	(void) free(strtab);

	/* Symbol and string tables modified, return */
	return F_OK;
}





/************************************************************************
*									*
*	Wrap():								*
*		o Read in the object header from the input file		*
*		o Copy the un-changed standard information - text, data	*
*		  and relocation entries to the output			*
*		o Read-Modify-Write the symbol and string tables	*
*		o Write an updated header to the output file		*
*									*
*	Currently, the size of the string table is stored in a 4 byte	*
*	header to the string table (just following the symbol table).	*
*	Since the only size parameter that changes in binary wrapping	*
*	is the string table size, the header of the object file does	*
*	not need to be changed. However, future enhancements to the	*
*	program may need to deal with relocation and program text	*
*	segments, in which case the header will change. Hence, the	*
*	header for the output file is written at the end of all changes	*
*									*
************************************************************************/

int	Wrap(inpFd, outFd, inpSym, outSym)
int	inpFd, outFd;
char	*inpSym, *outSym;
{
	/* Declare variables */
	struct exec	inpX, outX;


	/* Read the exec structure from the input file */
	if (ReadExec(inpFd, &inpX, inpObj) == F_ERR)
	{
		fprintf(stderr, "%s: unable to read header for %s\n",
			progName, inpObj);
		return F_ERR;
	}

	/* Copy input header to output header, need modification later */
	(void) memcpy((char *)&outX, (char *)&inpX, EXEC_SIZE);

	/* Copy text, data and relocation information as is to output */
	if (CopyStdInfo(inpFd, outFd, &inpX) == F_ERR)
	{
		fprintf(stderr, "%s: cannot copy text/data/reloc from %s to %s\n",
			progName, inpObj, outObj);
		return F_ERR;
	}

	/* Copy the symbol and string table info. after modification,
	   Also update the output exec structure as required */
	if (CopyTables(inpFd, outFd, inpSym, outSym, &inpX, &outX) == F_ERR)
	{
		fprintf(stderr, "%s: error modifying symbol/string tables\n",
			progName);
		return F_ERR;
	}

	/* Write the modified exec structure to output file */
	if (WriteExec(outFd, &outX, outObj) == F_ERR)
	{
		fprintf(stderr, "%s: unable to write header for %s\n",
			progName, outObj);
		return F_ERR;
	}

	/* Symbol table modified and updated, return */
	return F_OK;
}





/************************************************************************
*									*
*	main():								*
*		o Parse command line arguments, check usage		*
*		o Open input and output object files			*
*		o Call function to modify symbol entries		*
*		o Close object files					*
*									*
*	main() also initializes the global variable "progName" which	*
*	is the last pathname component of the executable file for this	*
*	program. The names of the input and output files are also set	*
*	for error message use by all the functions in the program.	*
*	Since the functions know if they are refering to the input or	*
*	output file, they can use the global variables for errors.	*
*									*
*	Strictly speaking, binary wrapping makes sense only for 	*
*	symbols that are text and external (non-text symbols are data	*
*	values are wrapping them is redundant; internal symbols are	*
*	useful only for debuggers, no external module can access them)	*
*									*
*	However, on SunOs, the "main" symbol in "crt0.o" is not defined	*
*	as text. In order to wrap startup modules, we should ignore the	*
*	type parameter for the symbols in these system libraries.	*
*	The program, by default, modifies all symbol strings - text and	*
*	data alike. The command line option "-t" can be specified to	*
*	force only external text symbols to be modified.		*
*									*
************************************************************************/

int	main(argc, argv)
int	argc;
char	**argv;
{
	/* Declare variables */
	char	*inpSym, *outSym;
	int	inpFd, outFd;


	/* Get program name */
	if ((progName = strrchr(*argv, '/')) == (char *)NULL)
		progName = *argv;
	else
		progName++;
	argv++; argc--;

	/* Check for text-only option */
	if ((argc != 0) && (strcmp(*argv, "-t") == 0))
	{
		textOnly = TRUE;
		argv++; argc--;
	}

	/* Check command line arguments */
	if (argc != NUM_ARGS)
	{
		/* Print usage, exit */
		Usage;
		return F_EXIT;
	}

	/* Initialize pointers to command line arguments */
	inpObj = *argv; argv++; argc--;
	outObj = *argv; argv++; argc--;
	inpSym = *argv; argv++; argc--;
	outSym = *argv; argv++; argc--;

	/* Open input and output objects */
	if ((inpFd = open(inpObj, O_RDONLY, 0)) < 0)
	{
		perror(inpObj);
		return F_ERR;
	}
	if ((outFd = open(outObj, (O_WRONLY|O_CREAT|O_TRUNC), OBJ_MODE)) < 0)
	{
		perror(outObj);
		return F_ERR;
	}

	/* Wrap the input file and write to output */
	if (Wrap(inpFd, outFd, inpSym, outSym, inpObj, outObj) == F_ERR)
	{
		/* Delete output file */
		fprintf(stderr, "%s: binary wrap for %s failed, deleting %s\n",
			progName, inpObj, outObj);
		if (unlink(outObj) < 0)
			perror(outObj);
		return F_ERR;
	}

	/* Close input and output objects */
	if (close(inpFd) < 0)
		perror(inpObj);
	if (close(outFd) < 0)
		perror(outObj);

	/* Program terminated */
	return F_OK;
}

